#!/usr/bin/env python2
# encoding: utf-8
import sys
from pyBusPirateLite.SPI import *

# based on http://ww1.microchip.com/downloads/en/DeviceDoc/22103a.pdf,
# gsupload, and nefariousdogooder's "GS transfer methods.txt"

# MCP23018 I2C device address
mcp_device_id = 0x20
mcp_read_byte = (mcp_device_id << 1) | 1
mcp_write_byte = (mcp_device_id << 1) | 0

# MCP23018 registers
IODIRA    = 0x00  # I/O direction: 0 write, 1 read
IODIRB    = 0x01
IOPOLA    = 0x02  # I/O polarity: 0 normal, 1 invert
IOPOLB    = 0x03
GPINTENA  = 0x04
GPINTENB  = 0x05
DEFVALA   = 0x06
DEFVALB   = 0x07
INTCONA   = 0x08
INTCONB   = 0x09
IOCON     = 0x0A # also 0x0B
GPPUA     = 0x0C # pullup: 0 off, 1 on
GPPUB     = 0x0D
INTFA     = 0x0E
INTFB     = 0x0F
INTCAPA   = 0x10
INTCAPB   = 0x11
GPIOA     = 0x12 # direct GPIO access
GPIOB     = 0x13
OLATA     = 0x14 # output latch
OLATB     = 0x15

def mcp_read(address):
  spi.CS_Low()
  ret = spi.bulk_trans(3,[mcp_read_byte, address, 0])
  spi.CS_High()
  return ord(ret[2])

def mcp_write(address, value):
  spi.CS_Low()
  spi.bulk_trans(3,[mcp_write_byte, address, value])
  spi.CS_High()

DEBUG = False

# my setup

def do_setup():
  global spi
  spi = SPI("/dev/bus_pirate", 115200)
  #spi = SPI("/dev/bus_pirate", 4000000)
  #spi = SPI("/dev/bus_pirate", 460800)

  print "Entering binmode: ",
  if spi.BBmode():
    print "OK."
  else:
    print "failed."
    sys.exit()

  print "Entering raw SPI mode: ",
  if spi.enter_SPI():
    print "OK."
  else:
    print "failed."
    sys.exit()
    
  print "Configuring SPI."
  #if spi.cfg_pins(PinCfg.POWER | PinCfg.PULLUPS | PinCfg.CS ) != 1:
  #if spi.cfg_pins(PinCfg.POWER | PinCfg.PULLUPS | PinCfg.CS | PinCfg.AUX) != 1:
  if spi.cfg_pins(PinCfg.POWER | PinCfg.CS ) != 1:
    print "Failed to set SPI peripherals."
    sys.exit()
  if spi.set_speed(SPISpeed._8MHZ) != 1:
    print "Failed to set SPI Speed."
    sys.exit()
  if spi.cfg_spi(SPICfg.OUT_TYPE | SPICfg.CLK_EDGE) != 1:
    print "Failed to set SPI configuration.";
    sys.exit()
  #spi.timeout(0.2)

  # MCP23018 setup

  # disable sequential operation
  # bit 5 = 1: sequential operation disabled
  # bit 7 = 0: bank = 0
  mcp_write(IOCON, 0x20)
  if do_read(IOCON) != 0x20:
    print "bad mojo"
    spi.resetBP()
    sys.exit()

  # A is output
  mcp_write(IODIRA, 0)
  do_read(IODIRA)
  # B is input
  mcp_write(IODIRB, 0xff)
  do_read(IODIRB)

  # pullups
  mcp_write(GPPUA, 0xff)
  mcp_write(GPPUB, 0)

def do_write(x, addr=OLATA):
  mcp_write(addr, x)
  if DEBUG: print "-> " + hex_string(mcp_read(addr))

def do_read(addr=GPIOB):
  x = mcp_read(addr)
  if DEBUG: print "<- " + hex_string(x)
  return x


# Gameshark stuff
def InitComms():
  print "Sync..."
  while True:
    # send two nibbles: 3 and 3, effectively '!'
    do_write(0x13)
    #spi.timeout(0.1)
    retval = do_read()>>4
    do_write(0);
    #spi.timeout(0.1)

    do_write(0x13)
    #spi.timeout(0.1)
    retval = (retval << 4) | (do_read()>>4)
    do_write(0)

    if retval == ord('g'): break

    # in case we are out of sync, grab one nibble
    do_write(0x13)
    #spi.timeout(0.1)
    retval = (retval << 4) | (do_read()>>4)
    do_write(0)

 
    print "retry"

  print "Handshake..."
  while True:
    if ReadWriteChar('G') != 'g': continue
    if ReadWriteChar('T') == 't': break
  print "OK"


def ReadWriteNibble(x, timeout=-1):
  do_write(x&0xf|0x10)
  spincount = 0

  while do_read() & 8 == 0:
    spincount += 1
    if timeout != -1:
      timeout -= 1
      if timeout == 0:
        print "timeout1"
        return 0

  retval = (do_read()&0xf0)>>4

  do_write(0)
  
  spincount = 0

  while do_read() & 8 == 8:
    spincount += 1
    if timeout != -1:
      timeout -= 1
      if timeout == 0:
        print "timeout2"
        return 0

  return retval

def ReadWriteByte(b, timeout=-1):
  return (ReadWriteNibble(b>>4, timeout)<<4)|ReadWriteNibble(b, timeout)

def ReadByte(timeout=-1):
  return ReadWriteByte(0, timeout)

def ReadWriteChar(c, timeout=-1):
  return chr(ReadWriteByte(ord(c), timeout))

def ReadChar(timeout=-1):
  return ReadWriteChar('\0', timeout)

def ReadWrite32(x, timeout=-1):
  return (ReadWriteNibble(x>>28,timeout)<<28)|\
         (ReadWriteNibble(x>>24,timeout)<<24)|\
         (ReadWriteNibble(x>>20,timeout)<<20)|\
         (ReadWriteNibble(x>>16,timeout)<<16)|\
         (ReadWriteNibble(x>>12,timeout)<<12)|\
         (ReadWriteNibble(x>> 8,timeout)<< 8)|\
         (ReadWriteNibble(x>> 4,timeout)<< 4)|\
         ReadWriteNibble(x,timeout)

def GSVersion():
  sig = [ReadWriteChar('f'), ReadChar(), ReadChar(), ReadChar()]

  strlen = ReadByte()
  version='' #hex_string(sig) + ': '
  for _ in range(strlen):
    version += ReadChar()

  return version


def Read(address, size):
  ReadWriteByte(1)
  ReadWrite32(address)
  ReadWrite32(size)

  b = ''
  for _ in range(size):
    b += ReadChar()
    print '.',
    sys.stdout.flush()

  print

  ReadWrite32(0)
  ReadWrite32(0)
  ReadWriteByte(0)

  return b

def Write(address, data):
  size = len(data)
  ReadWriteByte(2)
  ReadWrite32(address)
  ReadWrite32(size)

  for c in data:
    ReadWriteChar(c)
    print '.',
    sys.stdout.flush()

  print

  ReadWrite32(0)
  ReadWrite32(0)
  ReadWriteByte(0)

def Disconnect():
  InitComms()
  ReadWriteChar('d', 100)


# general debugging

def hex_string(s):
  if type(s) == int:
    return '%02x'%s
  else:
    return ''.join('%02x'%ord(c) for c in s)

def d(s):
  print(hex_string(s))


# 

if __name__ == '__main__':
  do_setup()
  
  # get the GS's attention
  InitComms()

  #print GSVersion()

  print "Reading"
  r = Read(0x80000000, 0x100)
  print hex_string(r)

  print "Writing"
  Write(0x80000000, r)
  Disconnect()

  mcp_write(GPIOA, 0)
  
  print "Reset Bus Pirate to user terminal: "
  if spi.resetBP():
    print "OK."
  else:
    print "failed."
    sys.exit()
    
